/*
Device: AtMega8
hfuse: 0xC9
lfuse: 0xE1

LCD on port D
Buttons on port B
32.768kHz crystal on TOSC
*/


/*	____________________Mega8___________________
	|Reset (PC6)		(ADC5/SCL) 	PC5|
	|PD0 (RXD)		(ADC4/SDA) 	PC4|
LCD:R/W	|PD1 (TXD)			(ADC3)	PC3|
LCD: E	|PD2 (Int0)			(ADC2)	PC2|
LCD: RS	|PD3 (Int2)			(ADC1)	PC1|
LCD:PB4	|PD4 (Xck/T0)			(ADC0)	PC0|
+5V	|VCC					GND|
Ground	|GND					AREF|
Crystal	|PB6 (Xtal1/Tosc1)			AVCC|
Crystal	|PB7 (Xtal2/Tosc2)		(SCK)	PB5| BTN: Down
LCD:PB5	|PD5 (T1)			(MISO)	PB4| BTN: Right
LCD:PB6	|PD6 (AIN0)		(MOSI/OC2)	PB3| BTN: Enter
LCD:PB7	|PD7 (AIN1)		(SS/OC1B)	PB2| BTN: Left
	|PB0 (ICP1)			(OC1A)	PB1| BTN: Up
	--------------------------------------------
*/
#define F_CPU 1000000UL

#include <avr/io.h>
#include <avr/interrupt.h>
#include "lcd.h"

typedef unsigned char   u8;
typedef signed short   s16;

//Store custom character information in program memory
//NOTE: Custom Characters for large number font was written by Xtinus
static const PROGMEM unsigned char BoxChar[] =
// Medium Char
{0x1f, 0x1f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, //   _
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1f, 0x1f, //         _
 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x1f, 0x1f, //  |_     _
 0x1f, 0x1f, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, //   _      |
 0x1f, 0x1f, 0x18, 0x18, 0x18, 0x18, 0x1f, 0x1f, //  |_     _
 0x1f, 0x1f, 0x03, 0x03, 0x03, 0x03, 0x1f, 0x1f, //  _      _|
 0x1f, 0x1f, 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, //   |
 0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x1f, 0x1f, //         _|
};

static const PROGMEM unsigned char BigNum[11][2][2] = { { {0x03, 0x06},    // 0
                                                          {0x02, 0x07}, }, // 0
                                                        { {0x06, 0x20},    // 1
                                                          {0x07, 0x01}, }, // 1
                                                        { {0x00, 0x05},    // 2
                                                          {0x04, 0x01}, }, // 2
                                                        { {0x00, 0x05},    // 3
                                                          {0x01, 0x07}, }, // 3
                                                        { {0x02, 0x07},    // 4
                                                          {0x20, 0x06}, }, // 4
                                                        { {0x04, 0x00},    // 5
                                                          {0x01, 0x05}, }, // 5
                                                        { {0x03, 0x00},    // 6
                                                          {0x04, 0x05}, }, // 6
                                                        { {0x00, 0x06},    // 7
                                                          {0x20, 0x06}, }, // 7
                                                        { {0x04, 0x05},    // 8
                                                          {0x02, 0x07}, }, // 8
                                                        { {0x04, 0x05},    // 9
                                                          {0x01, 0x07}, }, // 9
                                                        { {0x20, 0x20},    // clear
                                                          {0x20, 0x20}, }, // clear
};

//End of custom character info

//Debounce defines
#define KEY_PIN      PINB
#define KEY0      2 //Left
#define KEY1      1 //Up
#define KEY2      5 //Down
#define KEY3      4 //Right
#define KEY4      3 //Enter


//Used for RTC
#define RTC_compare_value 0x3D08  //Sets up the compare value for 1 second for real time clock (Internal osc 1MHz)

volatile unsigned int RTC_ticks = 0;
unsigned int RTC_mins = 0;
unsigned int RTC_hours = 0;
unsigned int old_ticks;
unsigned char format_24_hour = 1;

//Button Debounce Code
u8 digit; //Use to display a big number
u8 x_pos; //X position to display big character at

u8 key_state;            // debounced and inverted key state:
                         // bit = 1: key pressed
u8 key_press;            // key press detect

u8 get_key_press( u8 key_mask )
{
  cli();               // read and clear atomic !
  key_mask &= key_press;                        // read key(s)
  key_press ^= key_mask;                        // clear key(s)
  sei();
  return key_mask;
}
//End button debounce code (ISR for Timer0 also used for debouce)

void prog_custom_char(void)
{
  unsigned char i;
  lcd_command(_BV(LCD_CGRAM));  /* set CG RAM start address 0 */
  for(i=0; i<64; i++)
   {
     lcd_data(pgm_read_byte_near(&BoxChar[i]));  //Send custom characters from program memory to the LCD
   }
}

void disp_big_num(unsigned char digit, unsigned char x_pos)
{
  lcd_gotoxy(x_pos,0);                                 //Go to top line at position x-pos
  lcd_putc(pgm_read_byte_near(&BigNum[digit][0][0]));  //Write first custom character
  lcd_putc(pgm_read_byte_near(&BigNum[digit][0][1]));  //Write second custom character
  lcd_gotoxy(x_pos,1);                                 //Go to bottom line at posotion x-pos
  lcd_putc(pgm_read_byte_near(&BigNum[digit][1][0]));  //Write third custom character
  lcd_putc(pgm_read_byte_near(&BigNum[digit][1][1]));  //Write forth custom character
}

void disp_time(void)
{
  //LCD Spacing:
  //Hours-Tens at x=1
  //Hours-Ones at x=3
  //Colon      at x=5
  //Mins-Tens  at x=6
  //Mins-Ones  at x=8
  //Colon      at x=10
  //Secs-Tens  at x=11
  //Secs-Ones  at x=13

  //Display in 24 hour format
  if (format_24_hour)
  {
    //Hours tens:
    disp_big_num((RTC_hours/10),1);
    //Hours ones:
    disp_big_num((RTC_hours%10),3);
  }

  //Display in 12 hour format
  else
  {
    //Hours tens:
    char temp_hours = RTC_hours;
    lcd_gotoxy(15,1);  //Get ready to write am/pm
    if (temp_hours > 12) 
    {
      temp_hours -= 12;
      lcd_putc(0x50);  //Write the P for pm
    }
    else
    {
      if (temp_hours == 12) lcd_putc(0x50);  //Correct for 12 noon
      else lcd_putc(0x41);                  //Write teh A for am
    }
    if (temp_hours)
    {
      if (temp_hours/10) disp_big_num((temp_hours/10),1);
      else disp_big_num(10,1); //Clear the tens digit if not used.
      //Hours ones:
      disp_big_num((temp_hours%10),3);
    }
    else	//Correct for 12am (0 in 24 hour time)
    {
      disp_big_num(1,1);
      disp_big_num(2,3);
    }
  }

  //Colon:
  lcd_gotoxy(5,0);
  lcd_putc(0xa5);
  lcd_gotoxy(5,1);
  lcd_putc(0xa5);

  //Mins tens:
  disp_big_num((RTC_mins/10),6);
  //Mins Ones:
  disp_big_num((RTC_mins%10),8);

  //Colon:
  lcd_gotoxy(10,0);
  lcd_putc(0xa5);
  lcd_gotoxy(10,1);
  lcd_putc(0xa5);

  //Secs Tens:
  disp_big_num((old_ticks/10),11);
  //Secs Ones:
  disp_big_num((old_ticks%10),13);
}

void IncrementTime(void)
{
  if (++RTC_mins > 59) 
  {
    RTC_mins = 0;
    if (++RTC_hours > 23) RTC_hours = 0;
  }
}

void init_timers(void)
{
  //Timer0 for buttons
  TCCR0 |= 1<<CS02 | 1<<CS00;                //divide by 1024
  TIMSK |= 1<<TOIE0;                       //enable timer interrupt

  //Timer2 for RTC using 32.768kHz crystal on TOSC 1 and 2
  TCCR2 |= 1<<CS22 | 1<<CS20; //Set prescaler to 128 (causes overflow at 1 sec)
  ASSR |= 1<<AS2;             //Set Timer2 to asynchronous operation (used external crystal sparate from system clk)
  TIMSK |= 1<<TOIE2;          //Enable timer2 overflow interrupt

  /* - Using TOSC instead
  //Timer1 for RTC
  TCCR1B |= 1<<WGM12 | 1<<CS11 | 1<<CS10; //Turn on CTC with a prescaler of 64
  OCR1A = RTC_compare_value;              //Set compare value
  TIMSK |= 1<<OCIE1A;                     //enable timer compare interrupt
  sei();
  */
}

int main(void)
{
  lcd_init(LCD_DISP_ON);  //Initialize LCD with the cursor off
  prog_custom_char();     //Write custom characters to the LCD memory
  lcd_clrscr();           //Clear the LCD screen
  init_timers();          //Start debounce and RTC timers (Timer0 and Timer1 respectively)

  for (;;)
  {
    if (old_ticks != RTC_ticks)    //If 1 second has passed, update the display
    {
      if (old_ticks == 59) IncrementTime();
      old_ticks = RTC_ticks;
      disp_time();
    }
    if( get_key_press( 1<<KEY0 ))  //Increment hours
    {
      if(++RTC_hours > 23) RTC_hours = 0;
      disp_time();
    }
    if( get_key_press( 1<<KEY1 ))  //Increment minutes
    {
      if(++RTC_mins > 59) RTC_mins = 0;
      disp_time();
    }
    /*
    if( get_key_press( 1<<KEY2 ))
    {
      lcd_clrscr();
      lcd_gotoxy(3,0);
      lcd_puts("You Pressed");
      lcd_gotoxy(6,1);
      lcd_puts("Down");
    }
    */
    if( get_key_press( 1<<KEY3 ))  //Toggle 24 or 12 hour time
    {
      format_24_hour ^= 1;  //Toggle 24 hour time
      lcd_gotoxy(15,1);
      lcd_putc(0x20);       //Clear AM/PM marker
      disp_time();
    }
    if( get_key_press( 1<<KEY4 ))  //Reset seconds to zero (used to synchronize time)
    {
      //Reset the seconds counter.  Atomic!
      cli();
      RTC_ticks = 0;
      TCNT2 = 0; //Reset Timer2
      sei();
      disp_time();
    }
  }
}

//Debounce Overflow timer
ISR(TIMER0_OVF_vect)           // every 10ms
{
  static u8 ct0, ct1;
  u8 i;

  TCNT0 = (u8)(s16)-(XTAL / 1024 * 10e-3 + 0.5);   // preload for 10ms

  i = key_state ^ ~KEY_PIN;    // key changed ?
  ct0 = ~( ct0 & i );          // reset or count ct0
  ct1 = ct0 ^ (ct1 & i);       // reset or count ct1
  i &= ct0 & ct1;              // count until roll over ?
  key_state ^= i;              // then toggle debounced state
  key_press |= key_state & i;  // 0->1: key press detect
}

//RTC one second compare timer

ISR(TIMER2_OVF_vect)
{
  if (++RTC_ticks > 59) RTC_ticks = 0; //One minute has passed
}

/* using TOSC instead
ISR(TIMER1_COMPA_vect)
{
  if (++RTC_ticks > 59) RTC_ticks = 0; //One minute has passed
}
*/